#include "motor_thrd.h"
#include "motor_thrd_api.h"
#include "tof_thrd_api.h"
#include "sol_thrd_api.h"
#include "idc_assignment.h"
#include "motor_current_limit_api.h"

#include "SEGGER_RTT/SEGGER_RTT.h"

#define MOTOR_MODEOFOPERATION_MANUAL 1
#define MOTOR_MODEOFOPERATION_REACHDEST 2

int is12Open = 0;
static gpt_instance_ctrl_t *motorGpt[] = { &g_motor12_ctrl, &g_motor12_ctrl, &g_motor3_ctrl };
//static const timer_cfg_t *motorGptConfig[] = { &g_motor12_cfg, &g_motor12_cfg, &g_motor3_cfg };
static bsp_io_port_pin_t motorEnPin[] = { PIN_EN1, PIN_EN2, PIN_EN3 };
static bsp_io_port_pin_t motorDirPin[] = { PIN_DIR1, PIN_DIR2, PIN_DIR3 };
static gpt_io_pin_t motorGptChannel[] = { GPT_IO_PIN_GTIOCA, GPT_IO_PIN_GTIOCB, GPT_IO_PIN_GTIOCB };
static int motorDirectionInverse[] = { 0, 0, 0 };

static int modeOfOperation = MOTOR_MODEOFOPERATION_MANUAL;

static uint16_t targetTofX = 150;
static uint16_t targetTofY = 150;
static int directionX = MOTOR_DIRECTION_FORWARD;
static int directionY = MOTOR_DIRECTION_BACKWARD;
static TickType_t lastReachProgressPrint = 0;
static int ySpeed = 80;
static TickType_t lastSpeedIncrease = 0;

static void Motor_Init(int motorNumber) {
	fsp_err_t fStatus;

	if (motorNumber == 0 || motorNumber == 1) {
		if (!is12Open) {
			fStatus = R_GPT_Open(&g_motor12_ctrl, &g_motor12_cfg);
			is12Open = 1;
		} else {
			fStatus = FSP_SUCCESS;
		}
	} else {
		fStatus = R_GPT_Open(&g_motor3_ctrl, &g_motor3_cfg);
	}
	if (fStatus) {
		__BKPT(1);
	}

	Motor_TurnOff(motorNumber);

	fStatus = R_GPT_Start(motorGpt[motorNumber]);
	if (fStatus) {
		__BKPT(1);
	}
}

static void Motor_TurnOffInternal(int motorNumber) {
	R_IOPORT_PinWrite(&g_ioport_ctrl, motorEnPin[motorNumber], 1);
	R_GPT_OutputDisable(motorGpt[motorNumber], motorGptChannel[motorNumber]);
}

int Motor_TurnOff(int motorNumber) {
	if (modeOfOperation != MOTOR_MODEOFOPERATION_MANUAL) {
		return 1;
	}

	Motor_TurnOffInternal(motorNumber);
	return 0;
}

static void Motor_SetInternal(int motorNumber, int direction, int speed) {
	fsp_err_t fStatus;

	timer_ctrl_t *gpt = motorGpt[motorNumber];
	gpt_io_pin_t pwmChannel = motorGptChannel[motorNumber];
	bsp_io_port_pin_t dirPin = motorDirPin[motorNumber];
	bsp_io_port_pin_t enPin = motorEnPin[motorNumber];
	int inverse = motorDirectionInverse[motorNumber];

	if (direction == MOTOR_DIRECTION_FORWARD) {
		fStatus = R_IOPORT_PinWrite(&g_ioport_ctrl, dirPin, 0 ^ inverse);
	} else {
		fStatus = R_IOPORT_PinWrite(&g_ioport_ctrl, dirPin, 1 ^ inverse);
	}
	if (fStatus) {
		__BKPT(1);
	}

	fStatus = R_GPT_DutyCycleSet(gpt, (uint32_t)(speed * 50), pwmChannel);
	if (fStatus) {
		__BKPT(1);
	}

	fStatus = R_GPT_OutputEnable(gpt, pwmChannel);
	if (fStatus) {
		__BKPT(1);
	}

	fStatus = R_IOPORT_PinWrite(&g_ioport_ctrl, enPin, 0);
	if (fStatus) {
		__BKPT(1);
	}
}

int Motor_Set(int motorNumber, int direction, int speed) {
	if (modeOfOperation != MOTOR_MODEOFOPERATION_MANUAL) {
		return 1;
	}

	Motor_SetInternal(motorNumber, direction, speed);
	return 0;
}

int MotorSet_ReachDestination_Start(uint16_t tofXPos, uint16_t tofYPos) {
	if (modeOfOperation != MOTOR_MODEOFOPERATION_MANUAL) {
		return 1;
	}

	targetTofX = tofXPos;
	targetTofY = tofYPos;
	modeOfOperation = MOTOR_MODEOFOPERATION_REACHDEST;

	Motor_TurnOffInternal(MOTOR_X_NUMBER);
	Motor_TurnOffInternal(MOTOR_Y_NUMBER);

	ySpeed = 60;

	uint16_t actualTofX, actualTofY;
	TickType_t timestampX, timestampY;

	ToFManager_GetDistance(TOF_X_NUMBER, &timestampX, &actualTofX);
	ToFManager_GetDistance(TOF_Y_NUMBER, &timestampY, &actualTofY);

	TickType_t now = xTaskGetTickCount();

	if (now - timestampX > 1000 || now - timestampY > 100) {
		return 2;
	}

	if (targetTofX > actualTofX) {
		directionX = MOTOR_DIRECTION_FORWARD;
	} else {
		directionX = MOTOR_DIRECTION_BACKWARD;
	}

	if (targetTofY > actualTofY) {
		directionY = MOTOR_DIRECTION_FORWARD;
	} else {
		directionY = MOTOR_DIRECTION_BACKWARD;
	}

	lastReachProgressPrint = now;
	SEGGER_RTT_printf(0, "=============== REACH START ===============\r\n");
	SEGGER_RTT_printf(0, "TOFX: %d -> %d (direction: %s)\r\n", actualTofX, targetTofX, directionX == MOTOR_DIRECTION_FORWARD ? "FWD" : "BWD");
	SEGGER_RTT_printf(0, "TOFX: %d -> %d (direction: %s)\r\n", actualTofY, targetTofY, directionY == MOTOR_DIRECTION_FORWARD ? "FWD" : "BWD");
	SEGGER_RTT_printf(0, "===========================================\r\n");

	return 0;
}

static void MotorSet_ReachDestination_Process() {
	uint16_t actualTofX, actualTofY;
	TickType_t timestampX, timestampY;

	ToFManager_GetDistance(TOF_X_NUMBER, &timestampX, &actualTofX);
	ToFManager_GetDistance(TOF_Y_NUMBER, &timestampY, &actualTofY);

	TickType_t now = xTaskGetTickCount();

	if (now - timestampX > 1000 || now - timestampY > 100) {
		SEGGER_RTT_printf(0, "ToF failure while reaching destination. Waiting and retrying.\r\n");

		/*
		 modeOfOperation = MOTOR_MODEOFOPERATION_MANUAL;
		 */

		Motor_TurnOffInternal(MOTOR_X_NUMBER);
		Motor_TurnOffInternal(MOTOR_Y_NUMBER);

		if (ySpeed > 63) {
			ySpeed -= 3;
			SEGGER_RTT_printf(0, "Decreasing Y speed to %d.\r\n", ySpeed);
		}

		vTaskDelay(100);

		return;
	}

	if (now - lastSpeedIncrease > 200 && ySpeed < 100) {
		ySpeed += 1;
		lastSpeedIncrease = now;
		SEGGER_RTT_printf(0, "Increasing Y speed to %d.\r\n", ySpeed);
	}

	int iActualTofX = (int) actualTofX;
	int iActualTofY = (int) actualTofY;
	int iTargetTofX = (int) targetTofX;
	int iTargetTofY = (int) targetTofY;

	int remainingX, remainingY;
	int isMovingAny = 0;

	if (directionX == MOTOR_DIRECTION_FORWARD) {
		remainingX = iTargetTofX - iActualTofX;
	} else {
		remainingX = iActualTofX - iTargetTofX;
	}

	if (directionY == MOTOR_DIRECTION_FORWARD) {
		remainingY = iTargetTofY - iActualTofY;
	} else {
		remainingY = iActualTofY - iTargetTofY;
	}

	if (remainingX > 0) {
		Motor_SetInternal(MOTOR_X_NUMBER, directionX, 70);
		isMovingAny = 1;
	} else {
		Motor_TurnOffInternal(MOTOR_X_NUMBER);
	}

	if (remainingY > 0) {
		Motor_SetInternal(MOTOR_Y_NUMBER, directionY, 70);
		isMovingAny = 1;
	} else {
		Motor_TurnOffInternal(MOTOR_Y_NUMBER);
	}

	if (isMovingAny == 0) {
		modeOfOperation = MOTOR_MODEOFOPERATION_MANUAL;
		Solenoid_Trigger();
		SEGGER_RTT_printf(0, "=============== REACH SUCCESS ===============\r\n");
		return;
	}

	if (now - lastReachProgressPrint > 3000) {
		lastReachProgressPrint = now;
		SEGGER_RTT_printf(0, "=============== REACH PROGRESS ===============\r\n");
		SEGGER_RTT_printf(0, "TOFX: %d -> %d (direction: %s)\r\n", actualTofX, targetTofX, directionX == MOTOR_DIRECTION_FORWARD ? "FWD" : "BWD");
		SEGGER_RTT_printf(0, "TOFY: %d -> %d (direction: %s)\r\n", actualTofY, targetTofY, directionY == MOTOR_DIRECTION_FORWARD ? "FWD" : "BWD");
		SEGGER_RTT_printf(0, "==============================================\r\n");
	}
}

int MotorSet_ReachDestination_GetStatus() {
	return 0;
}

void motor_thrd_entry(void *pvParameters) {
	FSP_PARAMETER_NOT_USED(pvParameters);

	Motor_Init(MOTOR_SOLENOID_NUMBER);
	Motor_Init(MOTOR_X_NUMBER);
	Motor_Init(MOTOR_Y_NUMBER);

	motorDirectionInverse[MOTOR_X_NUMBER] = 1;

	while (1) {
		if (modeOfOperation == MOTOR_MODEOFOPERATION_REACHDEST) {
			MotorSet_ReachDestination_Process();
			vTaskDelay(1);
		} else {
			vTaskDelay(10);
		}
	}
}

