/*
 * COPYRIGHT (C) STMicroelectronics 2015. All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * STMicroelectronics ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with STMicroelectronics
 *
 * Programming Golden Rule: Keep it Simple!
 *
 */

/*!
 * \file   VL53L0X_platform.c
 * \brief  Code function defintions for Ewok Platform Layer
 *
 */


#include <stdio.h>    // sprintf(), vsnprintf(), printf()

#include "bsp_api.h"
#include "hal_data.h"
#include "r_iic_master.h"
#include "tof_thrd.h"

#include "vl53l0x_i2c_platform.h"
#include "vl53l0x_def.h"

#include "vl53l0x_platform_log.h"

#ifdef VL53L0X_LOG_ENABLE
#define trace_print(level, ...) trace_print_module_function(TRACE_MODULE_PLATFORM, level, TRACE_FUNCTION_NONE, ##__VA_ARGS__)
#define trace_i2c(...) trace_print_module_function(TRACE_MODULE_NONE, TRACE_LEVEL_NONE, TRACE_FUNCTION_I2C, ##__VA_ARGS__)
#endif


#define MIN_COMMS_VERSION_MAJOR     1
#define MIN_COMMS_VERSION_MINOR     8
#define MIN_COMMS_VERSION_BUILD     1
#define MIN_COMMS_VERSION_REVISION  0

#define STATUS_OK              0x00
#define STATUS_FAIL            0x01


static i2c_master_event_t i2cEvent = 0;

static void i2ccallback(i2c_master_callback_args_t* p_args) {
	i2cEvent = p_args->event;
}

static int i2c_waitforcompletion(i2c_master_event_t expectedEvent) {
	uint32_t timeout = 1000;
	while (i2cEvent == 0 && timeout--) {
		R_BSP_SoftwareDelay(1, BSP_DELAY_UNITS_MICROSECONDS);
		//__WFI();
	}

	if (i2cEvent == 0) {
		return 1;
	} else if (i2cEvent != expectedEvent) {
		R_IIC_MASTER_Abort(&g_i2c_master0_ctrl);
		return 2;
	}

	return 0;
}

int32_t VL53L0X_write_multi(uint8_t address, uint8_t index, uint8_t *pdata, int32_t count)
{
    int32_t status = STATUS_OK;

#ifdef VL53L0X_LOG_ENABLE
    int32_t i = 0;
    char value_as_str[VL53L0X_MAX_STRING_LENGTH_PLT];
    char *pvalue_as_str;

    pvalue_as_str =  value_as_str;

    for(i = 0 ; i < count ; i++)
    {
        sprintf(pvalue_as_str,"%02X", *(pdata+i));

        pvalue_as_str += 2;
    }
    trace_i2c("Write reg : 0x%04X, Val : 0x%s\n", index, value_as_str);
#endif

    fsp_err_t fStatus;
	uint8_t bufferToTransmit[64];
	bufferToTransmit[0] = index;

	if (count > 63) {
		return 55;
	}

	for (int32_t i = 0; i < count; i++) {
		bufferToTransmit[i + 1] = pdata[i];
	}

	fStatus = R_IIC_MASTER_SlaveAddressSet(&g_i2c_master0_ctrl, address, I2C_MASTER_ADDR_MODE_7BIT);
	if (fStatus) {
		return 1;
	}

	fStatus = R_IIC_MASTER_CallbackSet(&g_i2c_master0_ctrl, i2ccallback, NULL, NULL);
	if (fStatus) {
		return 1;
	}

	i2cEvent = 0;
	fStatus = R_IIC_MASTER_Write(&g_i2c_master0_ctrl, bufferToTransmit, (uint32_t)count + 1, 0);
	if (fStatus) {
		R_IIC_MASTER_Abort(&g_i2c_master0_ctrl);
		return 1;
	}
	status = i2c_waitforcompletion(I2C_MASTER_EVENT_TX_COMPLETE);
	if (status) {
		return status;
	}

    return 0;
}

int32_t VL53L0X_read_multi(uint8_t address, uint8_t index, uint8_t *pdata, int32_t count)
{
    int32_t status = STATUS_OK;

#ifdef VL53L0X_LOG_ENABLE
    int32_t      i = 0;
    char   value_as_str[VL53L0X_MAX_STRING_LENGTH_PLT];
    char *pvalue_as_str;
#endif

    fsp_err_t fStatus;

	fStatus = R_IIC_MASTER_SlaveAddressSet(&g_i2c_master0_ctrl, address, I2C_MASTER_ADDR_MODE_7BIT);
	if (fStatus) {
		return 1;
	}

	fStatus = R_IIC_MASTER_CallbackSet(&g_i2c_master0_ctrl, i2ccallback, NULL, NULL);
	if (fStatus) {
		return 1;
	}


	i2cEvent = 0;
	fStatus = R_IIC_MASTER_Write(&g_i2c_master0_ctrl, &index, 1, 1);
	if (fStatus) {
		R_IIC_MASTER_Abort(&g_i2c_master0_ctrl);
		return 11;
	}
	status = i2c_waitforcompletion(I2C_MASTER_EVENT_TX_COMPLETE);
	if (status) {
		return status;
	}

	i2cEvent = 0;
	fStatus = R_IIC_MASTER_Read(&g_i2c_master0_ctrl, pdata, (uint32_t)count, 0);
	if (fStatus) {
		R_IIC_MASTER_Abort(&g_i2c_master0_ctrl);
		return 22;
	}
	status = i2c_waitforcompletion(I2C_MASTER_EVENT_RX_COMPLETE);
	if (status) {
		return status;
	}


#ifdef VL53L0X_LOG_ENABLE
    // Build  value as string;
    pvalue_as_str =  value_as_str;

    for(i = 0 ; i < count ; i++)
    {
        sprintf(pvalue_as_str, "%02X", *(pdata+i));
        pvalue_as_str += 2;
    }

    trace_i2c("Read  reg : 0x%04X, Val : 0x%s\n", index, value_as_str);
#endif

    return 0;
}


int32_t VL53L0X_write_byte(uint8_t address, uint8_t index, uint8_t data)
{
    int32_t status = STATUS_OK;
    const int32_t cbyte_count = 1;

    status = VL53L0X_write_multi(address, index, &data, cbyte_count);

    return status;

}


int32_t VL53L0X_write_word(uint8_t address, uint8_t index, uint16_t data)
{
    int32_t status = STATUS_OK;

    uint8_t  buffer[BYTES_PER_WORD];

    // Split 16-bit word into MS and LS uint8_t
    buffer[0] = (uint8_t)(data >> 8);
    buffer[1] = (uint8_t)(data &  0x00FF);

    status = VL53L0X_write_multi(address, index, buffer, BYTES_PER_WORD);

    return status;

}


int32_t VL53L0X_write_dword(uint8_t address, uint8_t index, uint32_t data)
{
    int32_t status = STATUS_OK;
    uint8_t  buffer[BYTES_PER_DWORD];

    // Split 32-bit word into MS ... LS bytes
    buffer[0] = (uint8_t) (data >> 24);
    buffer[1] = (uint8_t)((data &  0x00FF0000) >> 16);
    buffer[2] = (uint8_t)((data &  0x0000FF00) >> 8);
    buffer[3] = (uint8_t) (data &  0x000000FF);

    status = VL53L0X_write_multi(address, index, buffer, BYTES_PER_DWORD);

    return status;

}


int32_t VL53L0X_read_byte(uint8_t address, uint8_t index, uint8_t *pdata)
{
    int32_t status = STATUS_OK;
    int32_t cbyte_count = 1;

    status = VL53L0X_read_multi(address, index, pdata, cbyte_count);

    return status;

}


int32_t VL53L0X_read_word(uint8_t address, uint8_t index, uint16_t *pdata)
{
    int32_t  status = STATUS_OK;
	uint8_t  buffer[BYTES_PER_WORD];

    status = VL53L0X_read_multi(address, index, buffer, BYTES_PER_WORD);
	*pdata = (uint16_t)(((uint16_t)buffer[0]<<8) + (uint16_t)buffer[1]);

    return status;

}

int32_t VL53L0X_read_dword(uint8_t address, uint8_t index, uint32_t *pdata)
{
    int32_t status = STATUS_OK;
	uint8_t  buffer[BYTES_PER_DWORD];

    status = VL53L0X_read_multi(address, index, buffer, BYTES_PER_DWORD);
    *pdata = (uint32_t)(((uint32_t)buffer[0]<<24) + ((uint32_t)buffer[1]<<16) + ((uint32_t)buffer[2]<<8) + (uint32_t)buffer[3]);

    return status;

}

int32_t VL53L0X_platform_wait_us(int32_t wait_us)
{
	R_BSP_SoftwareDelay(wait_us, BSP_DELAY_UNITS_MICROSECONDS);

#ifdef VL53L0X_LOG_ENABLE
    trace_i2c("Wait us : %6d\n", wait_us);
#endif

    return 0;

}


int32_t VL53L0X_wait_ms(int32_t wait_ms) {
	vTaskDelay(pdMS_TO_TICKS(wait_ms));

#ifdef VL53L0X_LOG_ENABLE
    trace_i2c("Wait ms : %6d\n", wait_ms);
#endif

    return 0;
}



int32_t VL53L0X_get_timer_frequency(int32_t *ptimer_freq_hz)
{
       *ptimer_freq_hz = 0;
       return STATUS_FAIL;
}


int32_t VL53L0X_get_timer_value(int32_t *ptimer_count)
{
       *ptimer_count = 0;
       return STATUS_FAIL;
}
