#include "tof_thrd.h"
#include "tof_thrd_api.h"
#include "vl53l0x_def.h"
#include "vl53l0x_api.h"
#include "SEGGER_RTT/SEGGER_RTT.h"
#include "r_iic_master.h"
#include "bsp_api.h"
#include "idc_assignment.h"

static int isFailed = 1; // default 1 triggers initialization on power-up
static VL53L0X_Dev_t tofSensors[3];
static int isUsed[3] = { 0, 0, 0 };
static uint8_t tofAddress[3] = { 0x20, 0x30, 0x40 };
static bsp_io_port_pin_t tofResetPins[3] = { BSP_IO_PORT_00_PIN_00, PIN_TOF_SHUT2, PIN_TOF_SHUT3 };

static uint16_t lastDistances[3] = { 0xFFFF, 0xFFFF, 0xFFFF };
static TickType_t lastMeasurementTime[3] = { 0, 0, 0 };
static int tofInvalidStatuses[3] = { 0, 0, 0 };

static int ToFManager_InitSensor(int tofNumber);

static int ToFManager_InitSensor(int tofNumber) {
	VL53L0X_Error vStatus;

	VL53L0X_Dev_t *tofSensor = tofSensors + tofNumber;

	tofSensor->I2cDevAddr = 0x52 >> 1;
	tofSensor->comms_type = 1;
	tofSensor->comms_speed_khz = 400;

	SEGGER_RTT_printf(0, "TOF #%d initialization started\r\n", tofNumber);

	vStatus = VL53L0X_SetDeviceAddress(tofSensor, tofAddress[tofNumber]);
	if (vStatus) {
		// We cant reset ToF 0 by HW pint because there is no reset line due to pinout constraints.
		// It may remained with already set address. We need to try reset it by SW.
		if (tofNumber == 0) {
			tofSensor->I2cDevAddr = tofAddress[0] >> 1;

			vStatus = VL53L0X_ResetDevice(tofSensor);
			if (vStatus) {
				SEGGER_RTT_printf(0, "Error while reseting TOF #%d\r\n", tofNumber);
				isFailed = 1;
				return 1;
			}

			// now we can try assigning address again
			tofSensor->I2cDevAddr = 0x52 >> 1;

			vStatus = VL53L0X_SetDeviceAddress(tofSensor, tofAddress[tofNumber]);
			if (vStatus) {
				SEGGER_RTT_printf(0, "Error while setting TOF #%d address\r\n", tofNumber);
				isFailed = 1;
				return 1;
			}

		} else {
			SEGGER_RTT_printf(0, "Error while setting TOF #%d address\r\n", tofNumber);
			isFailed = 1;
			return 1;
		}
	}

	tofSensor->I2cDevAddr = tofAddress[tofNumber] >> 1;

	VL53L0X_Version_t version;
	vStatus = VL53L0X_GetVersion(&version);
	if (vStatus) {
		SEGGER_RTT_printf(0, "Error while getting version of TOF #%d\r\n", tofNumber);
		isFailed = 1;
		return 1;
	}

	vStatus = VL53L0X_DataInit(tofSensor);
	if (vStatus) {
		SEGGER_RTT_printf(0, "Error while initializing data of TOF #%d\r\n", tofNumber);
		isFailed = 1;
		return 1;
	}

	VL53L0X_DeviceInfo_t deviceInfo;
	vStatus = VL53L0X_GetDeviceInfo(tofSensor, &deviceInfo);
	if (vStatus) {
		SEGGER_RTT_printf(0, "Error while getting TOF #%d device info.\r\n", tofNumber);
		isFailed = 1;
		return 1;
	}

	if ((deviceInfo.ProductRevisionMajor != 1) && (deviceInfo.ProductRevisionMinor != 1)) {
		SEGGER_RTT_printf(0, "Error expected cut 1.1 but found cut %d.%d\n", deviceInfo.ProductRevisionMajor, deviceInfo.ProductRevisionMinor);
		isFailed = 1;
		return 1;
	}

	SEGGER_RTT_printf(0, "============ TOF #%d VL53L0X_GetDeviceInfo ============\r\n", tofNumber);
	SEGGER_RTT_printf(0, "Device Name : %s\r\n", deviceInfo.Name);
	SEGGER_RTT_printf(0, "Device Type : %s\r\n", deviceInfo.Type);
	SEGGER_RTT_printf(0, "Device ID : %s\r\n", deviceInfo.ProductId);
	SEGGER_RTT_printf(0, "ProductRevisionMajor : %d\r\n", deviceInfo.ProductRevisionMajor);
	SEGGER_RTT_printf(0, "ProductRevisionMinor : %d\r\n", deviceInfo.ProductRevisionMinor);
	SEGGER_RTT_printf(0, "======================================================\r\n");

	vStatus = VL53L0X_StaticInit(tofSensor);
	if (vStatus) {
		SEGGER_RTT_printf(0, "Error while initializing TOF #%d\r\n", tofNumber);
		isFailed = 1;
		return 1;
	}

	uint8_t phaseCal;
	uint8_t vhvSettings;
	vStatus = VL53L0X_PerformRefCalibration(tofSensor, &vhvSettings, &phaseCal);
	if (vStatus) {
		SEGGER_RTT_printf(0, "Error while performing ref calibration of TOF #%d\r\n", tofNumber);
		isFailed = 1;
		return 1;
	}

	uint32_t refSpadCount;
	uint8_t isApertureSpads;
	vStatus = VL53L0X_PerformRefSpadManagement(tofSensor, &refSpadCount, &isApertureSpads);
	if (vStatus) {
		SEGGER_RTT_printf(0, "Error while performing ref spad management of TOF #%d\r\n", tofNumber);
		isFailed = 1;
		return 1;
	}

	vStatus = VL53L0X_SetDeviceMode(tofSensor, VL53L0X_DEVICEMODE_CONTINUOUS_RANGING);
	if (vStatus) {
		SEGGER_RTT_printf(0, "Error while setting TOF #%d device mode\r\n", tofNumber);
		isFailed = 1;
		return 1;
	}

	vStatus = VL53L0X_SetLimitCheckEnable(tofSensor, VL53L0X_CHECKENABLE_SIGMA_FINAL_RANGE, 1);
	if (vStatus) {
		SEGGER_RTT_printf(0, "Error while setting TOF #%d sigma final range limit check\r\n", tofNumber);
		isFailed = 1;
		return 1;
	}

	vStatus = VL53L0X_SetLimitCheckEnable(tofSensor, VL53L0X_CHECKENABLE_SIGNAL_RATE_FINAL_RANGE, 1);
	if (vStatus) {
		SEGGER_RTT_printf(0, "Error while setting TOF #%d final signal rate limit check\r\n", tofNumber);
		isFailed = 1;
		return 1;
	}

	vStatus = VL53L0X_SetLimitCheckEnable(tofSensor, VL53L0X_CHECKENABLE_RANGE_IGNORE_THRESHOLD, 1);
	if (vStatus) {
		SEGGER_RTT_printf(0, "Error while setting TOF #%d range ignore limit check\r\n", tofNumber);
		isFailed = 1;
		return 1;
	}

	vStatus = VL53L0X_SetLimitCheckValue(tofSensor, VL53L0X_CHECKENABLE_RANGE_IGNORE_THRESHOLD, (FixPoint1616_t) (1.5 * 0.023 * 65536));
	if (vStatus) {
		SEGGER_RTT_printf(0, "Error while setting TOF #%d range ingore threshold\r\n", tofNumber);
		isFailed = 1;
		return 1;
	}

	vStatus = VL53L0X_StartMeasurement(tofSensor);
	if (vStatus) {
		SEGGER_RTT_printf(0, "Error while starting TOF #%d\r\n", tofNumber);
		isFailed = 1;
		return 1;
	}

	SEGGER_RTT_printf(0, "TOF #%d initialization was sucessfull.\r\n", tofNumber);

	return 0;
}

static void ToFManager_ReadData(int tofNumber) {
	VL53L0X_Error vStatus;

	VL53L0X_Dev_t *tofSensor = tofSensors + tofNumber;
	uint8_t newDataReady;

	for (int i = 0; i < 5; i++) {
		vStatus = VL53L0X_GetMeasurementDataReady(tofSensor, &newDataReady);
		if (vStatus == 0) {
			break;
		}
		vTaskDelay(10);
	}
	if (vStatus) {
		SEGGER_RTT_printf(0, "Error while getting status of TOF #%d measurement data availability.\r\n", tofNumber);
		isFailed = 1;
		return;
	}

	TickType_t now = xTaskGetTickCount();

	if (newDataReady == 1) {
		// read data
		VL53L0X_RangingMeasurementData_t rangingData;
		vStatus = VL53L0X_GetRangingMeasurementData(tofSensor, &rangingData);
		if (vStatus) {
			SEGGER_RTT_printf(0, "Error while reading TOF #%d measurement data.\r\n", tofNumber);
			isFailed = 1;
			return;
		}

		if (rangingData.RangeStatus != 0) {
			// this can mean only warning not the error.
			// We ignore sample. In case of long-term issue timeout detection will trigger soon in the future.

			SEGGER_RTT_printf(0, "WARNING: TOF #%d provided sample with non-zero status code %d.\r\n", tofNumber, rangingData.RangeStatus);

			tofInvalidStatuses[tofNumber]++;

			if (tofInvalidStatuses[tofNumber] > 20) {
				isFailed = 1;
				return;
			}

			return;
		}

		lastDistances[tofNumber] = rangingData.RangeMilliMeter;
		lastMeasurementTime[tofNumber] = now;

		//SEGGER_RTT_printf(0, "TOF #%d range: %d\r\n", tofNumber, rangingData.RangeMilliMeter);
	} else {
		// check for time out
		if (now - lastMeasurementTime[tofNumber] > 10000) {
			SEGGER_RTT_printf(0, "ERROR: TOF #%d did not provided any data last 10000 ms.\r\n", tofNumber);
			isFailed = 1;
			return;
		}
	}
}

void ToFManager_GetDistance(int tofSensor, TickType_t *sampleTime, uint16_t *distance) {
	*sampleTime = lastMeasurementTime[tofSensor];
	*distance = lastDistances[tofSensor];
}

static void ToFManager_Initialize() {
	fsp_err_t fStatus;

	fStatus = R_IIC_MASTER_Close(&g_i2c_master0_ctrl);
	if (fStatus) {
		__BKPT(0);
	}

	fStatus = R_IIC_MASTER_Open(&g_i2c_master0_ctrl, &g_i2c_master0_cfg);
	if (fStatus) {
		__BKPT(0);
	}

	for (int i = 0; i < 20; i++) {
		g_i2c_master0_ctrl.p_reg->ICCR1_b.CLO = 1;
		while (g_i2c_master0_ctrl.p_reg->ICCR1_b.CLO) {
		}
	}

	// reset all sensors
	for (int i = 0; i < 3; i++) {
		// TOF1 cannot be reseted because xshut pin is always pulled up due to pinout utilization restrictions
		if (i != 0) {
			R_IOPORT_PinWrite(&g_ioport_ctrl, tofResetPins[i], 0);
		}
	}

	isFailed = 0;

	vTaskDelay(10);

	for (int i = 0; i < 3; i++) {
		if (isUsed[i]) {
			// TOF1 cannot be reseted because xshut pin is always pulled up due to pinout utilization restrictions
			if (i != 0) {
				R_IOPORT_PinWrite(&g_ioport_ctrl, tofResetPins[i], 1);
			}
			vTaskDelay(10);
			ToFManager_InitSensor(i);

			if (isFailed) {
				break;
			}
		}
	}
}

void tof_thrd_entry(void *pvParameters) {
	FSP_PARAMETER_NOT_USED(pvParameters);

	fsp_err_t sts = R_IIC_MASTER_Open(&g_i2c_master0_ctrl, &g_i2c_master0_cfg);
	if (sts) {
		__BKPT(0);
	}

	for (int i = 0; i < 20; i++) {
		g_i2c_master0_ctrl.p_reg->ICCR1_b.CLO = 1;
		while (g_i2c_master0_ctrl.p_reg->ICCR1_b.CLO) {
		}
	}

	isUsed[TOF_X_NUMBER] = 1;
	isUsed[TOF_Y_NUMBER] = 1;

	while (1) {
		if (isFailed) {
			ToFManager_Initialize();
		} else {
			for (int i = 0; i < 3; i++) {
				if (isUsed[i]) {
					ToFManager_ReadData(i);

					if (isFailed) {
						break;
					}
				}
			}
		}

		vTaskDelay(50);
	}
}
